/*
 *   fdn.c  - a feedback delay network (reverb tail)
 *   using a housholder reflection feedback matrix (In - 2/n 11T)
 *   Copyright (c) 2000-2003 by Tom Schouten
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* TODO: CLEAN UP THIS MESS
 
 add delay time generation code
 add prime calculation routine (for prime delay line lengths)
 add more diffuse feedback matrix (hadamard)
 check filtering code
 
 */

#include "ext.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "ext_obex.h"
#include "z_dsp.h"
#include <math.h>


#define FDN_MIN_DECAY_TIME .01

/*
 
 #define NBPRIMES
 int prime[NBPRIMES];
 
 static int isprime(int n)
 {
 int i=1;
 int d,m,p;
 while(1){
 p = prime[i++];
 m = n % p;
 if (m == 0) return 0; // it is a prime
 d = n / p;
 if (d < p) return 1;  // it is not a prime
 }
 }
 
 static int initprimes(void)
 {
 int i, curprime;
 prime[0] = 1;
 prime[1] = 2;
 curprime = 3;
 
 for(i=2; i<NBPRIMES; i++){
 while (!isprime(curprime)) curprime++;
 prime[i] = curprime;
 //printf("%d, ", curprime);
 curprime++;
 }
 printf("\n");
 return 0;
 }
 */


//static int find_nearest_prime(int n){ return n;}

typedef struct fdnctl
{
	int c_order;      /* veelvoud van 4 */
	int c_maxorder;
	float c_leak;
	float c_input;
	float c_output;
	float *c_buf;
	float *c_gain_in;
	float *c_gain_state;
	float c_timehigh;
	float c_timelow;
	int *c_tap;       /* cirular feed: N+1 pointers: 1 read, (N-1)r/w, 1 write */ 
	float *c_length;  /* delay lengths in ms */
	int c_bufsize;
	float c_fsample;
	float *c_vector[2];
	float *c_vectorbuffer;
	int c_curvector;
} t_fdnctl;

typedef struct fdn
{
	t_object x_obj;
	float x_f;
	t_fdnctl x_ctl;
} t_fdn;



void fdn_order(t_fdn *x, int order){
	if (order > x->x_ctl.c_maxorder) {
		post("fdn: this should not happen (panic!) order %d "
			 "is larger than maxorder %d:", 
			 order, x->x_ctl.c_maxorder ); 
		exit(1);
	}
	x->x_ctl.c_order = order;
	x->x_ctl.c_leak = -2./ order;
	x->x_ctl.c_input = 1./ sqrt(order); //????????????????
	
}

void fdn_print(t_fdn *x)
{
	int i;
	fprintf(stderr, "fdn: delay coefficients (ms)\n");
	for (i=0;i<x->x_ctl.c_order;i++) {
		fprintf(stderr, "%f ", x->x_ctl.c_length[i]);
	}
	fprintf(stderr, "\n");
	
}


 void fdn_reset(t_fdn *x)
{
	int i;
	if (x->x_ctl.c_buf) 
		memset(x->x_ctl.c_buf, 0, x->x_ctl.c_bufsize 
			   * sizeof(float));
	if (x->x_ctl.c_vectorbuffer) 
		memset(x->x_ctl.c_vectorbuffer,
			   0, x->x_ctl.c_maxorder * 2 * sizeof(float));
}




 int *fdn_perform(int *w)
{
	
	
	float *in      = (float *)(w[3]);
	float *outr    = (float *)(w[4]);
	float *outl    = (float *)(w[5]);
	t_fdnctl *ctl    = (t_fdnctl *)(w[1]);
	int n          = (int)(w[2]);
	float input    = ctl->c_input;
	float output   = ctl->c_output;
	float *gain_in    = ctl->c_gain_in;
	float *gain_state = ctl->c_gain_state;
	int order      = ctl->c_order;
	int *tap       = ctl->c_tap;
	float *buf     = ctl->c_buf;
	int mask       = ctl->c_bufsize - 1;
	
	int i,j;
	float x,y,v,left,right,z;
	float filt_in, filt_last;
	
	float *cvec, *lvec;
	
	float save;
	
	
	for(i=0;i<n;i++){
		
		x = *in++;
		y = 0;
		left = 0;
		right = 0;
		
		/* get temporary vector buffers */
		cvec = ctl->c_vector[ctl->c_curvector];
		lvec = ctl->c_vector[ctl->c_curvector ^ 1];
		ctl->c_curvector ^= 1;
		
		
		/* read input vector + get sum and left/right output*/
		for(j=0;j<order;) 
		{
			z = buf[tap[j]];
			cvec[j] = z;
			y     += z;
			left  += z;
			right += z;
			j++;
			
			z = buf[tap[j]];
			cvec[j] = z;
			y     += z;
			left  -= z;
			right += z;
			j++;
			
			z = buf[tap[j]];
			cvec[j] = z;
			y     += z;
			left  += z;
			right -= z;
			j++;
			
			z = buf[tap[j]];
			cvec[j] = z;
			y     += z;
			left  -= z;
			right -= z;
			j++;
		}	
		
		
		/* write output */
		*outl++ = left;
		*outr++ = right;
		
		/* y == leak to all inputs */
		y *=  ctl->c_leak;
		
		
		
		/* perform feedback */
		/* todo: decouple feedback & permutation */
		save = cvec[0];
		for (j=0; j<order-1; j++){
			cvec[j] = cvec[j+1] + y + x; 
		}
		cvec[order-1] = save + y + x;
		
		
		
		/* apply gain + store result vector in delay lines + increment taps*/
		tap[0] = (tap[0]+1)&mask;
		for(j=0;j<order;j++) {
			save = gain_in[j] * cvec[j] + gain_state[j] * lvec[j];
			save = IS_DENORMAL(save) ? 0 : save;
			cvec[j] = save;
			buf[tap[j+1]] = save;
			tap[j+1] = (tap[j+1] + 1) & mask;
		}	
	}
	
    
	return (w+6);
}

void fdn_dsp(t_fdn *x, t_signal **sp, short *count)
{
	
	x->x_ctl.c_fsample = sp[0]->s_sr;
	dsp_add(fdn_perform, 
			5, 
			&x->x_ctl, 
			sp[0]->s_n, 
			sp[0]->s_vec, 
			sp[1]->s_vec, 
			sp[2]->s_vec);
}                                  
	
	
 void fdn_free(t_fdn *x)
{
	if ( x->x_ctl.c_tap) free( x->x_ctl.c_tap);
	if ( x->x_ctl.c_length) free( x->x_ctl.c_length);
	if ( x->x_ctl.c_gain_in) free( x->x_ctl.c_gain_in);
	if ( x->x_ctl.c_gain_state) free( x->x_ctl.c_gain_state);
	if ( x->x_ctl.c_buf) free ( x->x_ctl.c_buf);
	if ( x->x_ctl.c_vectorbuffer) free ( x->x_ctl.c_vectorbuffer );
}


/*
 
 each delay line is filtered with a first order iir filter:
 (gl: dc gain, gh: ny gain)
 
 H(z) = 2 gl gh / (gl + gh - z^-1 (gl - gh))
 
 this results in the difference equation
 
 yk = (2 gl gh ) / (gl + gh) x + (gl - gh) / (gl + gh) yk-1
 
 */


 void fdn_time(t_fdn *x, float timelow, float timehigh){
	float elow, ehigh;
	int i;
	float gainlow, gainhigh, gainscale;
	
	if (timelow < FDN_MIN_DECAY_TIME) timelow = FDN_MIN_DECAY_TIME;
	if (timehigh < FDN_MIN_DECAY_TIME) timehigh = FDN_MIN_DECAY_TIME;
	
	elow = -.003 / (timelow);
	ehigh = -.003 / (timehigh);
	
	/* setup gains */
	for(i=0;i<x->x_ctl.c_order;i++){
		gainlow = pow(10, elow * (x->x_ctl.c_length[i]));
		gainhigh = pow(10, ehigh * (x->x_ctl.c_length[i]));
		gainscale = 1.0f / (gainlow + gainhigh);
		x->x_ctl.c_gain_in[i]    = 2.0f * gainlow * gainhigh * gainscale;
		x->x_ctl.c_gain_state[i] = (gainlow - gainhigh) * gainscale;
	}
	x->x_ctl.c_timehigh = timehigh;
	x->x_ctl.c_timelow = timelow;
}

 void fdn_updatedamping(t_fdn *x)
{
	fdn_time(x, x->x_ctl.c_timelow, x->x_ctl.c_timehigh);
}

 void fdn_timelow(t_fdn *x, float f){
	x->x_ctl.c_timelow = fabs(f);
	fdn_updatedamping(x);
}
 void fdn_timehigh(t_fdn *x, float f){
	x->x_ctl.c_timehigh = fabs(f);
	fdn_updatedamping(x);
}


 void fdn_setupdelayline(t_fdn *x){
	int sum, t, n;
	int mask = x->x_ctl.c_bufsize - 1;
	int start =  x->x_ctl.c_tap[0];
	int *tap = x->x_ctl.c_tap;
	float *length = x->x_ctl.c_length;
	float scale = sys_getsr() * .001f;
	
	sum = 0;
	tap[0] = (start & mask);
	for (t=1; t<= x->x_ctl.c_order; t++){
		sum += (int)(length[t-1] * scale);
		tap[t]=(start+sum)&mask;
	}
	
	if (sum > mask){ 
		post("fdn: warning: not enough delay memory, behaviour "
			 "is undefined (this could lead to instability...)");
	}
    
	
}

 void fdn_list (t_fdn *x,  t_symbol *s, int argc, t_atom *argv){
	int i;
	float l;
	int sum=0;
	
	int order = argc & 0xfffffffc;
	
	if (order < 4) return;
	if (order > x->x_ctl.c_maxorder) return;
	
	
	fdn_order(x, order);
	for(i=0; i<order; i++)
		if (argv[i].a_type == A_FLOAT) x->x_ctl.c_length[i] = argv[i].a_w.w_float;
	
	fdn_setupdelayline(x);
	fdn_updatedamping(x);
}

 void fdn_linear(t_fdn *x, float forder, float min, float max)
{
    int order = ((int)forder) & 0xfffffffc;
    float length, inc;
    int i;
	
    if (order < 4) return;
    if (order > x->x_ctl.c_maxorder) return;
    if (min <= 0) return;
    if (max <= 0) return;
	
    inc = (max - min) / (float)(order - 1);
    length = min;
    
    for (i=0; i<order; i++){
		x->x_ctl.c_length[i] = length;
		length += inc;
    }
	
    fdn_order(x, order);
    fdn_setupdelayline(x);
    fdn_updatedamping(x);
} 



 void fdn_exponential(t_fdn *x, float forder, float min, float max)
{
    int order = ((int)forder) & 0xfffffffc;
    float length, inc;
    int i;
	
    if (order < 4) return;
    if (order > x->x_ctl.c_maxorder) return;
    if (min <= 0) return;
    if (max <= 0) return;
	
    inc = pow (max / min, 1.0f / ((float)(order - 1)));
    length = min;
    
    for (i=0; i<order; i++){
		x->x_ctl.c_length[i] = length;
		length *= inc;
    }
	
    fdn_order(x, order);
    fdn_setupdelayline(x);
    fdn_updatedamping(x);
} 





void *fdn_class;

 void *fdn_new(float maxiorder, float maxibufsize) //a partir de la, ya du boulot
{
	int order = maxiorder;
	int bufround;
	t_fdn *x;
	x = (t_fdn *) object_alloc((t_class *)fdn_class);
	float scale = sys_getsr() * .001f;
	int bufsize = (int)(scale * maxibufsize);
	
/* 	pd :
 x->sig_out = outlet_new(&x->x_obj, &s_signal);
 x->bang_out = outlet_new(&x->x_obj, &s_bang); // essai
 x->f_out = outlet_new(&x->x_obj, &s_float); // essai
 x->f_out2 = outlet_new(&x->x_obj, &s_float); // essai
	max :
 x->intout = intout((t_object *)x); // V		
 x->intout2 = intout((t_object *)x); // V
 x->bangout = bangout((t_object *)x); // V
 outlet_new(x, "signal");
*/	
	
//	inlet_new(&x->x_obj, &x->x_obj.ob_pd, "float", "timelow");  
//	inlet_new(&x->x_obj, &x->x_obj.ob_pd, "float", "timehigh");  
//	outlet_new(&x->x_obj, "signal"); 
//	outlet_new(&x->x_obj, "signal"); 

	intin((t_object *)x, "float");
	intin((t_object *)x, "float");
	
	/* init data */
	if (order < 4) order = 8;
	if (bufsize < 64) bufsize = 65536;
	
	bufround = 1;
	while (bufround < bufsize) bufround *= 2;
	bufsize = bufround;
	
	
	post("fdn: maximum nb of delay lines %d, total buffer "
		 "size %d samples (%f seconds)", 
		 order, bufsize, ((float)bufsize) / sys_getsr());
	float
	
	x->x_ctl.c_maxorder = order;
	x->x_ctl.c_buf = (float *)malloc(sizeof(float) * bufsize);
	x->x_ctl.c_bufsize = bufsize;
	x->x_ctl.c_fsample = sys_getsr();
	x->x_ctl.c_tap = (int *)malloc((order + 1) * sizeof(int)); 
	x->x_ctl.c_length = (float *)malloc(order * sizeof(int)); 
	x->x_ctl.c_gain_in = (float *)malloc(order * sizeof(float));
	x->x_ctl.c_gain_state = (float *)malloc(order * sizeof(float));
	x->x_ctl.c_vectorbuffer = (float *)malloc(order * 2 * sizeof(float));
	memset(x->x_ctl.c_vectorbuffer, 0, order * 2 * sizeof(float));
	x->x_ctl.c_curvector = 0;
	x->x_ctl.c_vector[0] = &x->x_ctl.c_vectorbuffer[0];
	x->x_ctl.c_vector[1] = &x->x_ctl.c_vectorbuffer[order];
	
	/* preset */
	fdn_order(x,8);
	x->x_ctl.c_length[0]= 29.0f;
	x->x_ctl.c_length[1]= 31.0f;
	x->x_ctl.c_length[2]= 37.0f;
	x->x_ctl.c_length[3]= 67.0f;
	x->x_ctl.c_length[4]= 82.0f;
	x->x_ctl.c_length[5]= 110.0f;
	x->x_ctl.c_length[6]= 172.0f;
	x->x_ctl.c_length[7]= 211.0f;
	fdn_setupdelayline(x);
	fdn_time(x, 4, 1);
	
	/* reset delay memory to zero */
	fdn_reset(x);
	
	
	return (void *)x;
}


int main(void)
{
	//post("fdn~ v0.1");
	t_class *c;
	
    c = class_new("fdn~", (method)fdn_new,
						  (method)fdn_free, sizeof(t_fdn), 0, A_DEFFLOAT, A_DEFFLOAT, 0);
	
//    CLASS_MAINSIGNALIN(c, t_fdn, x_f); // ???
//	CLASS_MAINSIGNALIN
//	
//	CLASS_MAINSIGNALIN(<class_name>, <class_data>, <f>);
//	
//	The macro CLASS_MAINSIGNALIN declares, that the class will use signal-inlets.
//	
//	The first macro-argument is a pointer to the signal-class. The second argument is the type of the class-dataspace. The third argument is a (dummy-)floatingpoint-variable of the dataspace, that is needed to automatically convert ``float''-messages into signals if no signal is present at the signal-inlet.
//		
//		No ``float''-methods can be used for signal-inlets, that are created this way. 

	//class_addmethod(abs_class, (t_method)nullfn, &s_signal, A_NULL);
    class_addmethod(c, (method)fdn_print, "print", 0);
    class_addmethod(c, (method)fdn_reset, "reset", 0);
    class_addmethod(c, (method)fdn_timehigh,
					"timehigh", A_DEFFLOAT, 0);
    class_addmethod(c, (method)fdn_timelow,
					"timelow", A_DEFFLOAT, 0);
    class_addmethod(c, (method)fdn_list, "lines", A_GIMME, 0);
    class_addmethod(c, (method)fdn_dsp, "dsp", 0); 
    class_addmethod(c, (method)fdn_linear,
					"linear", A_FLOAT, A_FLOAT, A_FLOAT, 0); 
    class_addmethod(c, (method)fdn_exponential,
					"exponential", A_FLOAT, A_FLOAT, A_FLOAT, 0); 
	class_dspinit(c);
	class_register(CLASS_BOX, c);
	fdn_class = c;
	return 0;
}

